import logging
from functools import lru_cache

from brownie import web3
from multicall import Call, Multicall
from y.contracts import Contract, has_methods
from y.decorators import log
from y.exceptions import ContractNotVerified
from ypricemagic.utils.multicall import fetch_multicall
from ypricemagic.utils.raw_calls import _decimals

logger = logging.getLogger(__name__)

# NOTE: Yearn and Yearn-like

@log(logger)
@lru_cache(maxsize=None)
def is_yearn_vault(token):
    logger.debug(f'Checking `is_yearn_vault({token})')
    # Yearn-like contracts can use these formats
    result = any([
        has_methods(token, ['pricePerShare()(uint)','getPricePerShare()(uint)','getPricePerFullShare()(uint)','getSharesToUnderlying()(uint)'], any),
        has_methods(token, ['exchangeRate()(uint)','underlying()(address)']),
    ])

    # pricePerShare can revert if totalSupply == 0, which would cause `has_methods` to return `False`,
    # but it might still be a vault. This section will correct `result` for problematic vaults.
    if result is False:
        try: 
            contract = Contract(token)
            result = any([
                hasattr(contract,'pricePerShare'),
                hasattr(contract,'getPricePerShare'),
                hasattr(contract,'getPricePerFullShare'),
                hasattr(contract,'getSharesToUnderlying'),
            ])
        except ContractNotVerified: pass

    logger.debug(f'`is_yearn_vault({token})` returns `{result}`')
    return result


@log(logger)
def get_price(token, block=None):
    # v1 vaults use getPricePerFullShare scaled to 18 decimals
    # v2 vaults use pricePerShare scaled to underlying token decimals
    # yearnish clones use all sorts of other things, we gotchu covered

    share_price_methods = ['pricePerShare()(uint)','getPricePerShare()(uint)','getPricePerFullShare()(uint)','getSharesToUnderlying()(uint)','exchangeRate()(uint)']
    calls = [Call(token, [method], [[method, None]]) for method in share_price_methods]
    results = [result for result in Multicall(calls, _w3=web3, block_id=block, require_success=False)().values() if result is not None]
    assert len(results) in [1,0], 'Something is going wrong in yearn vault calculations. Must debug'
    if len(results) == 1: share_price = results[0]

    underlying_methods = ['token()(address)','underlying()(address)','native()(address)','want()(address)']
    calls = [Call(token, [method], [[method, None]]) for method in underlying_methods]
    results = [result for result in Multicall(calls, _w3=web3, block_id=block, require_success=False)().values() if result is not None]
    assert len(results) in [1,0], 'Something is going wrong in yearn vault calculations. Must debug'
    if len(results) == 1: underlying = results[0]

    decimals = _decimals(token)

    # saving for later
    '''
    vault = Contract(token)
    if hasattr(vault, 'pricePerShare'):
        share_price, underlying, decimals = fetch_multicall(
            [vault, 'pricePerShare'],
            [vault, 'token'],
            [vault, 'decimals'],
            block=block
        )
    elif hasattr(vault, 'getPricePerFullShare') and hasattr(vault, 'token'):
        share_price, underlying = fetch_multicall(
            [vault, 'getPricePerFullShare'],
            [vault, 'token'],
            block=block
        )
        decimals = 18
    elif hasattr(vault, 'getPricePerFullShare') and hasattr(vault, 'underlying'):
        share_price, underlying, decimals = fetch_multicall(
            [vault, 'getPricePerFullShare'],
            [vault, 'underlying'],
            [vault, 'decimals'],
            block=block
        )
    elif hasattr(vault, 'getPricePerFullShare') and hasattr(vault, 'native'):
        share_price, underlying, decimals = fetch_multicall(
            [vault, 'getPricePerFullShare'],
            [vault, 'native'],
            [vault, 'decimals'],
            block=block
        )
    elif hasattr(vault, 'getPricePerFullShare') and hasattr(vault, 'want'):
        share_price, underlying, decimals = fetch_multicall(
            [vault, 'getPricePerFullShare'],
            [vault, 'want'],
            [vault, 'decimals'],
            block=block
        )
    elif vault.__dict__['_build']['contractName'] == 'BeefyVaultV6Matic':
        share_price, underlying, decimals = fetch_multicall(
            [vault, 'getPricePerFullShare'],
            [vault, 'wmatic'],
            [vault, 'decimals'],
            block=block
        )
    elif vault.__dict__['_build']['contractName'] == 'BeefyVenusVaultBNB':
        share_price, underlying, decimals = fetch_multicall(
            [vault, 'getPricePerFullShare'],
            [vault, 'wbnb'],
            [vault, 'decimals'],
            block=block
        )
    elif hasattr(vault, 'getPricePerShare'):
        share_price, underlying, decimals = fetch_multicall(
            [vault, 'getPricePerShare'],
            [vault, 'token'],
            [vault, 'decimals'],
            block=block
        )
    elif hasattr(vault,'exchangeRate'):
        share_price, underlying, decimals = fetch_multicall(
            [vault,'exchangeRate'],
            [vault,'underlying'],
            [vault,'decimals'],
            block=block
        )
    elif hasattr(vault,'getSharesToUnderlying'):
        underlying, decimals = fetch_multicall(
            [vault,'token'],
            [vault,'decimals'],
            block=block
        )
        share_price = vault.getSharesToUnderlying(10 ** decimals,block_identifier=block)
    '''
    ''' might need this later for goofy L1s w/o multicall2        
    else:
        if hasattr(vault, 'pricePerShare'):
            share_price = vault.pricePerShare(block_identifier = block)
            underlying = vault.token(block_identifier = block)
            decimals = vault.decimals(block_identifier = block)
        elif hasattr(vault, 'getPricePerFullShare') and hasattr(vault, 'token'):
            share_price = vault.getPricePerFullShare(block_identifier = block)
            underlying = vault.token(block_identifier = block)
            decimals = 18
        elif hasattr(vault, 'getPricePerFullShare') and hasattr(vault, 'underlying'):
            share_price = vault.getPricePerFullShare(block_identifier = block)
            underlying = vault.underlying(block_identifier = block)
            decimals = vault.decimals(block_identifier = block)
        elif hasattr(vault, 'getPricePerFullShare') and hasattr(vault, 'want'):
            share_price = vault.getPricePerFullShare(block_identifier = block)
            underlying = vault.want(block_identifier = block)
            decimals = vault.decimals(block_identifier = block)
        elif hasattr(vault, 'getPricePerFullShare') and hasattr(vault, 'native'):
            share_price = vault.getPricePerFullShare(block_identifier=block)
            underlying = vault.native(block_identifier=block)
            decimals = vault.decimals(block_identifier=block)
        elif hasattr(vault, 'getPricePerShare'):
            share_price = vault.getPricePerShare(block_identifier = block)
            underlying = vault.token(block_identifier = block)
            decimals = vault.decimals(block_identifier = block)
            '''

    #try:
    price = [share_price / 10 ** decimals, underlying]
    #except TypeError: # when getPricePerShare() reverts due to divide by zero
    #    price = [1, underlying]
    #except UnboundLocalError: # not supported
    #    price = None
    
    if price: logger.debug("yearn -> %s", price)
    return price



import logging
from functools import lru_cache

from brownie import ZERO_ADDRESS, web3
from multicall import Call
from y.contracts import has_method
from y.decorators import log
from y.erc20 import decimals, totalSupplyReadable
from y.exceptions import call_reverted
from ypricemagic import magic
from ypricemagic.utils.multicall import multicall_balanceOf
from ypricemagic.utils.raw_calls import _decimals, raw_call

logger = logging.getLogger(__name__)


@log(logger)
@lru_cache
def is_pie(address):
    result = has_method(address, "getCap()(uint)")
    return result

@log(logger)
def get_price(pie_address: str, block=None):
    price = tvl(pie_address, block) / totalSupplyReadable(pie_address, block)
    return price

@log(logger)
def get_tokens(pie_address: str, block: int = None):
    return Call(pie_address, ['getTokens()(address[])'], [['tokens',None]], _w3=web3, block_id=block)()['tokens']

@log(logger)
def get_bpool(pie_address: str, block: int = None):
    try:
        bpool = raw_call(pie_address, 'getBPool()', output='address', block=block)
        return bpool if bpool != ZERO_ADDRESS else pie_address
    except Exception as e:
        if call_reverted(e): return pie_address
        else: raise

@log(logger)
def tvl(pie_address: str, block: int = None):
    tokens = get_tokens(pie_address, block)
    pool = get_bpool(pie_address, block)
    token_balances = multicall_balanceOf(tokens, pool, block=block)
    token_decimals = decimals(tokens, block)
    token_balances = [bal / 10 ** decimal for bal, decimal in zip(token_balances, token_decimals)]
    prices = magic.get_prices(tokens, block, silent=True, dop=1)
    return sum(bal * price for bal, price in zip(token_balances, prices))

# currently unused:
PPROXYPAUSABLE = [{"inputs":[],"stateMutability":"nonpayable","type":"constructor"},{"stateMutability":"payable","type":"fallback"},{"inputs":[{"internalType":"address","name":"_value","type":"address"}],"name":"addressToBytes32","outputs":[{"internalType":"bytes32","name":"","type":"bytes32"}],"stateMutability":"pure","type":"function"},{"inputs":[{"internalType":"bytes32","name":"_value","type":"bytes32"}],"name":"bytes32ToAddress","outputs":[{"internalType":"address","name":"","type":"address"}],"stateMutability":"pure","type":"function"},{"inputs":[{"internalType":"bytes32","name":"_value","type":"bytes32"}],"name":"bytes32ToString","outputs":[{"internalType":"string","name":"","type":"string"}],"stateMutability":"pure","type":"function"},{"inputs":[],"name":"getImplementation","outputs":[{"internalType":"address","name":"","type":"address"}],"stateMutability":"view","type":"function"},{"inputs":[],"name":"getPaused","outputs":[{"internalType":"bool","name":"","type":"bool"}],"stateMutability":"view","type":"function"},{"inputs":[],"name":"getPauzer","outputs":[{"internalType":"address","name":"","type":"address"}],"stateMutability":"view","type":"function"},{"inputs":[],"name":"getProxyOwner","outputs":[{"internalType":"address","name":"","type":"address"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"bytes32","name":"_key","type":"bytes32"}],"name":"readAddress","outputs":[{"internalType":"address","name":"","type":"address"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"bytes32","name":"_key","type":"bytes32"}],"name":"readBool","outputs":[{"internalType":"bool","name":"","type":"bool"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"bytes32","name":"_key","type":"bytes32"}],"name":"readString","outputs":[{"internalType":"string","name":"","type":"string"}],"stateMutability":"view","type":"function"},{"inputs":[],"name":"renouncePauzer","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"address","name":"_newImplementation","type":"address"}],"name":"setImplementation","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"bool","name":"_value","type":"bool"}],"name":"setPaused","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"address","name":"_newPauzer","type":"address"}],"name":"setPauzer","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"address","name":"_newOwner","type":"address"}],"name":"setProxyOwner","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"bytes32","name":"_key","type":"bytes32"}],"name":"storageRead","outputs":[{"internalType":"bytes32","name":"","type":"bytes32"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"string","name":"_value","type":"string"}],"name":"stringToBytes32","outputs":[{"internalType":"bytes32","name":"result","type":"bytes32"}],"stateMutability":"pure","type":"function"}]
